/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana;

import java.lang.Thread.UncaughtExceptionHandler;
import java.text.MessageFormat;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import com.bitctrl.util.jar.JarTools;
import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.ObjectTimeSpecification;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.UmfelddatenobjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UmfeldDatenMessStelle;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UmfeldDatenSensor;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.VerkehrsobjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenTeilSegment;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList.Argument;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Die Datenverteilerapplikation <i>Straßensubsegmentanalyse</i>. Segment 5
 * Intelligente Analyseverfahren, SWE 5.2 Straßensubsegmentanalyse
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public final class SubSegAna
implements StandardApplication, UncaughtExceptionHandler {

	/** Verbindung zum Datenverteiler. */
	private static ClientDavInterface verbindung;

	/**
	 * Gibt die Datenverteilerverbindung der Applikation zurück.
	 *
	 * @return die Datenverteilerapplikation.
	 */
	public static ClientDavInterface getVerbindung() {
		return verbindung;
	}

	/**
	 * Einsprungmethode der Applikation <i>Straßensubsegmentanalyse</i>.
	 *
	 * @param args
	 *            die Kommandozeilenargumente
	 * @see #parseArguments(ArgumentList)
	 */
	public static void main(final String[] args) {
		StandardApplicationRunner.run(new SubSegAna(), args);
	}

	/** Logger. */
	private Debug log;

	/** Enthält die PIDs der zu verarbeitenden Systemobjekte. */
	private String[] argSystemobjekte;

	/** Maximale Anzahl zu cachender Zyklen für die Regelbasis. */
	private int argMaxZyklen;

	/**
	 * Initialisiert die Applikation.
	 */
	private SubSegAna() {
		JarTools.printVersionInfo(SubSegAna.class);
	}

	/**
	 * Initialisierung und Start der Straßensubsegmentanalyse und deren
	 * Submodule.
	 */
	@Override
	public void initialize(final ClientDavInterface connection)
			throws Exception {
		MessageSender.getInstance()
		.setApplicationLabel(getClass().getSimpleName() + "_"
				+ connection.getLocalApplicationObject().getId());

		final Set<SystemObjekt> systemobjekte;
		final Set<MessQuerschnittAllgemein> mqMenge;
		final Set<UmfeldDatenSensor> ufdsMenge;

		// Lokale Variablen initialisieren
		verbindung = connection;

		// Zu verwaltende Objekte bestimmen
		LogTools.log(log, SubSegAnaMsg.InfoBestimmeSystemobjekte);
		ObjektFactory.getInstanz().setVerbindung(verbindung);
		ObjektFactory.getInstanz().registerFactory(new VerkehrsobjektFactory());
		ObjektFactory.getInstanz()
		.registerFactory(new UmfelddatenobjektFactory());
		if (argSystemobjekte == null) {
			final DataModel modell = connection.getDataModel();
			final SystemObject[] objekte = modell
					.getObjects(null,
							Collections.singletonList(modell
									.getType("typ.konfigurationsBereich")), //$NON-NLS-1$
					ObjectTimeSpecification.valid())
									.toArray(new SystemObject[0]);

			argSystemobjekte = new String[objekte.length];
			for (int i = 0; i < argSystemobjekte.length; i++) {
				argSystemobjekte[i] = objekte[i].getPid();
			}
		}
		systemobjekte = new HashSet<>(ObjektFactory.getInstanz()
				.bestimmeModellobjekte(argSystemobjekte));

		// Messquerschnitte zu Straßenteilsegmenten bestimmen
		mqMenge = new HashSet<>();
		for (final SystemObjekt so : systemobjekte) {
			if (so instanceof StrassenTeilSegment) {
				StrassenTeilSegment sts;

				sts = (StrassenTeilSegment) so;
				mqMenge.addAll(sts.getMessQuerschnitte());
			}
		}
		systemobjekte.addAll(mqMenge);

		// Umfelddatensensoren zu Umfelddatenmessstellen bestimmen
		ufdsMenge = new HashSet<>();
		for (final SystemObjekt so : systemobjekte) {
			if (so instanceof UmfeldDatenMessStelle) {
				UmfeldDatenMessStelle ufdms;

				ufdms = (UmfeldDatenMessStelle) so;
				ufdsMenge.addAll(ufdms.getUmfelddatensensoren());
			}
		}
		systemobjekte.addAll(ufdsMenge);

		// Submodule starten
		starteSubmodule(systemobjekte);

		// nicht mehr benötigte lokale Variablen freigeben
		argSystemobjekte = null;

		LogTools.log(log, SubSegAnaMsg.InfoSubSegAnaBereit);
	}

	/**
	 * Übernimmt die Werte der Aufrufparameter. Folgende Parameter werden von
	 * der Applikation ausgewertet:
	 * <ul>
	 * <li>{@code -objekte}: Eine mit Komma getrennte Liste von PID von
	 * Systemobjekten (optional, Standard: alle)</li>
	 * <li>{@code -maxZyklen}: Die maximale Anzahl zu cachende Zyklen von
	 * Fuzzy-Werten (optional, Standard: 2, erlaubt 1 .. 10)</li>
	 * </ul>
	 */
	@Override
	public void parseArguments(final ArgumentList argumentList)
			throws Exception {
		Argument arg;

		Thread.setDefaultUncaughtExceptionHandler(this);
		log = Debug.getLogger();

		// Zu verarbeitende Systemobjekte bestimmen
		if (argumentList.hasArgument("-objekte")) { //$NON-NLS-1$
			arg = argumentList.fetchArgument("-objekte"); //$NON-NLS-1$
			argSystemobjekte = arg.asNonEmptyString().split(","); //$NON-NLS-1$
			for (int i = 0; i < argSystemobjekte.length; i++) {
				argSystemobjekte[i] = argSystemobjekte[i].trim();
			}
		}

		// Maximale Anzahl zu cachender Zyklen bestimmen
		arg = argumentList.fetchArgument("-maxZyklen=2"); //$NON-NLS-1$
		argMaxZyklen = arg.intValueBetween(1, 10);
	}

	/**
	 * Gibt die Exception im Log und als Betriebsmeldung aus und beendet
	 * anschließend die Applikation.
	 */
	@Override
	public void uncaughtException(final Thread t, final Throwable e) {
		LogTools.log(log, SubSegAnaMsg.ErrorUnbekannterFehler, t,
				LoggerTools.getStackTrace(e));
		if (log == null) {
			System.err.println(MessageFormat.format(
					SubSegAnaMsg.ErrorUnbekannterFehler.toString(), t,
					LoggerTools.getStackTrace(e)));
		}
		System.exit(-1);
	}

	/**
	 * Startet die einzelnen Submodule der Applikation und verbindet sie.
	 *
	 * @param systemobjekte
	 *            Zu verwaltende Systemobjekte
	 */
	private void starteSubmodule(final Collection<SystemObjekt> systemobjekte) {
		MesswertEmpfaenger empfaenger;
		Fuzzyfizierung fuzzy;
		Aggregation aggregation;

		// Parameterverwaltung starten
		@SuppressWarnings("unused")
		final ParameterVerwaltung parameterVerwaltung = new ParameterVerwaltung(
				systemobjekte);

		// Messwertempfänger starten
		empfaenger = new MesswertEmpfaenger(systemobjekte);

		// Fuzzyfizierung starten
		fuzzy = new Fuzzyfizierung(systemobjekte);

		// Aggregation starten
		aggregation = new Aggregation(systemobjekte);

		// Regelbasis starten
		@SuppressWarnings("unused")
		final Regelbasis regelbasis = new Regelbasis(systemobjekte,
				argMaxZyklen);

		// Listener registrieren
		empfaenger.addMesswertListener(fuzzy);
		empfaenger.addMesswertListener(aggregation);
		aggregation.addMesswertListener(fuzzy);
	}

}
