/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.2 Straßensubsegmentanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.subsegana.subsegana;

import java.util.Collection;

import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzyVariable;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.FuzzyWert;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.OdFuzzyWert;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.PdFuzzySets;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdmsGlaetteFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsHelligkeitFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsNiederschlagsIntensitaetFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsSichtWeiteFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsWindGeschwindigkeitMittelWertFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.umfelddaten.OdUfdsWindRichtungFuzzy;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdAnteilLkw;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdBemessungsDichte;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdGeschwindkeit;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdVerkehrsStaerkeStunde;
import de.bsvrz.iav.fuzzylib.fuzzylib.modell.verkehr.OdVerkehrsUnruhe;
import de.bsvrz.iav.subsegana.subsegana.events.MesswertEvent;
import de.bsvrz.iav.subsegana.subsegana.events.MesswertListener;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.AnmeldeException;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.MesswertDatum;
import de.bsvrz.sys.funclib.bitctrl.modell.ParameterDatensatz;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsHelligkeit;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsNiederschlagsIntensitaet;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsSichtWeite;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsWindGeschwindigkeitMittelWert;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UfdsWindRichtung;
import de.bsvrz.sys.funclib.bitctrl.modell.umfelddaten.objekte.UmfeldDatenMessStelle;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenTeilSegment;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Submodul Fuzzyfizierung. Empfängt Analysewerte vom Datenverteiler,
 * fuzzyfiziert diese und publiziert die fuzzyfizierten Werte mit Hilfe des
 * Datenverteilers.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Fuzzyfizierung implements MesswertListener {

	/** Logger. */
	private final Debug log = Debug.getLogger();

	/**
	 * Meldet sich für die Fuzzy-Werte der Systemobjekte als Quelle an.
	 *
	 * @param objekte
	 *            Menge von Systemobjekten
	 */
	Fuzzyfizierung(final Collection<SystemObjekt> objekte) {
		int anzahlObjekte, anzahlFuzzyWerte;

		LogTools.log(log, SubSegAnaMsg.InfoStarteFuzzyfizierung);

		// Als Quelle der Fuzzy-Werte anmelden
		anzahlObjekte = 0;
		anzahlFuzzyWerte = 0;
		for (final SystemObjekt so : objekte) {
			if (so instanceof MessQuerschnittAllgemein) {
				anzahlObjekte++;
				anzahlFuzzyWerte += anmeldenSender(so, OdAnteilLkw.class);
				anzahlFuzzyWerte += anmeldenSender(so,
						OdBemessungsDichte.class);
				anzahlFuzzyWerte += anmeldenSender(so, OdGeschwindkeit.class);
				anzahlFuzzyWerte += anmeldenSender(so,
						OdVerkehrsStaerkeStunde.class);
				anzahlFuzzyWerte += anmeldenSender(so, OdVerkehrsUnruhe.class);
			} else if (so instanceof StrassenTeilSegment) {
				final StrassenTeilSegment sts = (StrassenTeilSegment) so;
				if (sts.getMessQuerschnitte().isEmpty()) {
					// STS ohne MQ können ignoriert werden
					continue;
				}

				anzahlObjekte++;
				anzahlFuzzyWerte += anmeldenSender(so, OdAnteilLkw.class);
				anzahlFuzzyWerte += anmeldenSender(so,
						OdBemessungsDichte.class);
				anzahlFuzzyWerte += anmeldenSender(so, OdGeschwindkeit.class);
				anzahlFuzzyWerte += anmeldenSender(so,
						OdVerkehrsStaerkeStunde.class);
				anzahlFuzzyWerte += anmeldenSender(so, OdVerkehrsUnruhe.class);
			} else if (so instanceof UfdsHelligkeit) {
				anzahlObjekte++;
				anzahlFuzzyWerte += anmeldenSender(so,
						OdUfdsHelligkeitFuzzy.class);
			} else if (so instanceof UfdsNiederschlagsIntensitaet) {
				anzahlObjekte++;
				anzahlFuzzyWerte += anmeldenSender(so,
						OdUfdsNiederschlagsIntensitaetFuzzy.class);
			} else if (so instanceof UfdsWindGeschwindigkeitMittelWert) {
				anzahlObjekte++;
				anzahlFuzzyWerte += anmeldenSender(so,
						OdUfdsWindGeschwindigkeitMittelWertFuzzy.class);
			} else if (so instanceof UfdsSichtWeite) {
				anzahlObjekte++;
				anzahlFuzzyWerte += anmeldenSender(so,
						OdUfdsSichtWeiteFuzzy.class);
			} else if (so instanceof UfdsWindRichtung) {
				anzahlObjekte++;
				anzahlFuzzyWerte += anmeldenSender(so,
						OdUfdsWindRichtungFuzzy.class);
			} else if (so instanceof UmfeldDatenMessStelle) {
				anzahlObjekte++;
				anzahlFuzzyWerte += anmeldenSender(so,
						OdUfdmsGlaetteFuzzy.class);
			} else {
				LogTools.log(log,
						new BetriebsmeldungDaten(so.getSystemObject()),
						SubSegAnaMsg.FinerIgnoriereUnbekanntesSystemObject, so);
			}
		}

		LogTools.log(log, SubSegAnaMsg.ConfigFuzzyfizierungBereit,
				anzahlObjekte, anzahlFuzzyWerte);
	}

	/**
	 * Nimmt die Analysedaten entgegen, fuzzyfiziert sie und schickt die
	 * Fuzzy-Werte weiter.
	 */
	@Override
	public void messwertAktualisiert(final MesswertEvent event) {
		SystemObjekt so;
		MesswertDatum datum;

		so = event.getObjekt();
		datum = event.getDatum();
		for (final String w : datum.getWerte()) {
			final Number messwert = datum.getWert(w);
			boolean gefunden = false;

			for (final ParameterDatensatz<?> pd : so.getParameterDatensatz()) {
				if (pd instanceof PdFuzzySets) {
					final PdFuzzySets fs = (PdFuzzySets) pd;
					if (w.equals(fs.getName())) {
						if (!fs.getDatum().isValid()) {
							// Ungültiges Datum wird ignoriert
							LogTools.log(log,
									new BetriebsmeldungDaten(
											so.getSystemObject()),
									SubSegAnaMsg.WarningUngueltigerFuzzyParameter,
									so, fs.getAttributGruppe(),
											event.getDatensatz().getAttributGruppe());
							continue;
						}

						FuzzyVariable fv;
						OdFuzzyWert fw;
						FuzzyWert wert;

						// Fuzzyfizieren
						fv = null; // Zustand "nicht ermittelbar"
						if (messwert != null) {
							fv = fs.getDatum().getLinguistischeVariable()
									.fuzzyfiziere(messwert.intValue());
						}

						// Als Datum verpacken ...
						fw = so.getOnlineDatensatz(fs.getFuzzyWert());
						wert = fw.erzeugeDatum();
						wert.setZeitstempel(datum.getZeitstempel());
						wert.setFuzzyWert(fv);

						// ... und verschicken
						try {
							LogTools.log(log,
									new BetriebsmeldungDaten(
											so.getSystemObject()),
									SubSegAnaMsg.FineSendeFuzzyWert, so,
											fw.getAttributGruppe(), wert);
							fw.sendeDaten(OdFuzzyWert.Aspekte.Fuzzy.getAspekt(),
									wert);

						} catch (final DatensendeException ex) {
							LogTools.log(log,
									new BetriebsmeldungDaten(
											so.getSystemObject()),
									SubSegAnaMsg.ErrorBeimSendenFuzzyWert, so,
									fw.getAttributGruppe(), wert, ex);
						}
						gefunden = true;
					}
				}
			}
			if (!gefunden) {
				LogTools.log(log,
						new BetriebsmeldungDaten(so.getSystemObject()),
						SubSegAnaMsg.FinerIgnoriereUnbekanntenMesswert, so, w);
			}
		}
	}

	/**
	 * Meldet sich als Quelle eines Datensatzes eines Systemobjekts an.
	 *
	 * @param so
	 *            ein Systemobjekt.
	 * @param datensatz
	 *            der zu sendende Datensatz.
	 * @return 1 beim erfolgreichen Anmelden, sonst 0.
	 */
	private int anmeldenSender(final SystemObjekt so,
			final Class<? extends OdFuzzyWert> datensatz) {
		OdFuzzyWert od;

		od = so.getOnlineDatensatz(datensatz);
		od.setQuelle(OdFuzzyWert.Aspekte.Fuzzy.getAspekt(), true);

		LogTools.log(log, new BetriebsmeldungDaten(so.getSystemObject()),
				SubSegAnaMsg.FinerAnmeldenQuelleFuzzyWert, so,
				od.getAttributGruppe());
		try {
			so.getOnlineDatensatz(datensatz)
			.anmeldenSender(OdFuzzyWert.Aspekte.Fuzzy.getAspekt());

			return 1;
		} catch (final AnmeldeException ex) {
			LogTools.log(log, new BetriebsmeldungDaten(so.getSystemObject()),
					SubSegAnaMsg.ErrorBeimAnmeldenQuelleFuzzyWert, so,
					od.getAttributGruppe(), ex);

			return 0;
		}
	}
}
